from random import randint

from mirakuru import TCPExecutor


class TestGRPCServer:
    def __init__(self, manage_py, params=None):
        if params is None:
            params = {}
        params.setdefault('--port', 50000 + randint(0, 10000))
        self.port = params.get('--port')
        self.manage_py = manage_py
        self.process = TCPExecutor(
            ['python', self.manage_py, 'grpcserver'] + list(self.flat_params(params)),
            host='localhost', port=self.port
        )

    @classmethod
    def flat_params(cls, dict_):
        for k, v in dict_.items():
            yield str(k)
            if str(v) != "":
                yield str(v)

    def addr(self):
        return "localhost:%s" % self.port

    def start(self):
        self.process.start()

    def stop(self):
        self.process.stop()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from shutil import rmtree
from setuptools import setup, Command


here = os.path.abspath(os.path.dirname(__file__))

about = {}
# Load the package's __version__.py module as a dictionary.
with open(os.path.join(here, 'django_grpc', '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        return_status = os.WEXITSTATUS(os.system('%s setup.py sdist bdist_wheel --universal' % sys.executable))
        if return_status:
            self.status("Failed to build the package!")
            sys.exit(return_status)

        self.status("Checking distribution...")
        return_status = os.WEXITSTATUS(os.system("twine check dist/*"))
        if return_status:
            self.status("Failed to validate the package!")
            sys.exit(return_status)

        self.status('Uploading the package to PyPI via Twine...')
        return_status = os.WEXITSTATUS(os.system('twine upload dist/*'))
        if return_status:
            self.status("Failed to upload the package!")
            sys.exit(return_status)

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


readme = open('README.md').read()
# history = open('HISTORY.md').read().replace('.. :changelog:', '')

setup(
    name='django-grpc',
    version=about['__version__'],
    description="""Easy Django based gRPC service""",
    long_description=readme,  # + '\n\n' + history,
    long_description_content_type="text/markdown",
    author='Stan Misiurev',
    author_email='smisiurev@gmail.com',
    url='https://github.com/gluk-w/django-grpc',
    packages=[
        'django_grpc',
        'django_grpc_testtools',
    ],
    include_package_data=True,
    install_requires=['setuptools'],
    license="MIT",
    zip_safe=False,
    keywords='django-grpc',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    cmdclass={
        'upload': UploadCommand,
    },
)



import datetime

from django.core.management.base import BaseCommand
from django.utils import autoreload

from django_grpc.utils import create_server, extract_handlers


class Command(BaseCommand):
    help = 'Run gRPC server'

    def add_arguments(self, parser):
        parser.add_argument('--max_workers', type=int, help="Number of workers")
        parser.add_argument('--port', type=int, default=50051, help="Port number to listen")
        parser.add_argument('--autoreload', action='store_true', default=False)
        parser.add_argument('--list-handlers', action='store_true', default=False, help="Print all registered endpoints")

    def handle(self, *args, **options):
        if options['autoreload'] is True:
            self.stdout.write("ATTENTION! Autoreload is enabled!")
            if hasattr(autoreload, "run_with_reloader"):
                # Django 2.2. and above
                autoreload.run_with_reloader(self._serve, **options)
            else:
                # Before Django 2.2.
                autoreload.main(self._serve, None, options)
        else:
            self._serve(**options)

    def _serve(self, max_workers, port, *args, **kwargs):
        autoreload.raise_last_exception()
        self.stdout.write("Starting server at %s" % datetime.datetime.now())

        server = create_server(max_workers, port)
        server.start()

        self.stdout.write("Server is listening port %s" % port)

        if kwargs['list_handlers'] is True:
            self.stdout.write("Registered handlers:")
            for handler in extract_handlers(server):
                self.stdout.write("* %s" % handler)

        server.wait_for_termination()



import hashlib
import time
from functools import reduce
from typing import Callable, List, Optional, Tuple, Union

from functools import wraps

import grpc
from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured


def get_current_time_window(time_period: int):
    """Calculates end of the current time window for a given time period."""
    ts = int(time.time())
    if time_period == 1:
        return ts
    w = ts - (ts % time_period)
    if w < ts:
        return w + time_period
    return w


def create_cache_key(group: str, values: List[str], time_window: int):
    """Creates storage key for specified group and values of current time_period."""
    parts = [group, ".".join(values), str(time_window)]
    return hashlib.md5(''.join(parts).encode('utf-8')).hexdigest()


def get_keys_values(request, context, keys: List[Union[str, Callable]]) -> List[str]:
    values = []
    for key in keys:
        # User provided function to get key from request and context
        if callable(key):
            values.append(key(request, context))
        # Gets keys value from gRPCs message
        elif key.startswith("request:"):
            _, fields_dot_path = key.split(":")
            value = str(reduce(lambda msg, field: getattr(msg, field, ""), fields_dot_path.split("."), request))
            values.append(value)
        # Gets keys value from gRPCs metadata
        elif key.startswith("metadata:"):
            _, metadata_key = key.split(":")
            value = dict(context.invocation_metadata()).get(metadata_key, "") or ""
            values.append(value)

    return values


def save_call(cache_key: str, time_period: int) -> int:
    """Saves current call and returns number of calls for given key."""
    cache_name = getattr(settings, 'RATELIMIT_USE_CACHE', 'default')
    cache = caches[cache_name]

    count = 1
    # Extend the expiration time by a few seconds to avoid misses.
    added = cache.add(cache_key, count, time_period + 3)
    if not added:
        try:
            # memcached will throw a ValueError if server unavailable or key does not exist.
            count = cache.incr(cache_key)
        except ValueError:
            pass

    return count


def record_call(
    rpc,
    request,
    context,
    time_period: int,
    group: str = None,
    keys: List[Union[str, Callable]] = None,
) -> Tuple[int, int]:
    """Records call and returns current calls count.

    :param rpc: gRPCs procedure
    :param context: gRPCs context
    :param request: gRPCs message
    :param time_period: Time period per which limit calls
    :param group: Name of a group of ratelimits to count together.
        Basically to share the same ratelimit across one or more RPCs.
    :param keys: Client/s identifier. Like `group` but is used to group calls based on gRPCs message and context.
        For example by clients user-agent, ip and other requests information.
    :returns: Current count of calls and time left in seconds until enf of time period
    """
    if group is None:
        # By default group will be RPCs class' and methods name
        group = rpc.__qualname__
    if keys is None:
        keys = []

    if time_period <= 0:
        raise ImproperlyConfigured('time_period must be greater than 0')

    time_window = get_current_time_window(time_period)
    values = get_keys_values(request, context, keys)

    cache_key = create_cache_key(group, values, time_window)
    count = save_call(cache_key, time_period)

    time_left = time_window - int(time.time())
    return count, time_left


def ratelimit(max_calls: int, time_period: int, group: Optional[str] = None, keys: List[Union[str, Callable]] = None):
    """
    :param max_calls: Max number of calls in specified `time_period`.
    :param time_period: Time period in seconds per which limit `max_calls`.
    :param group: Name of a group of ratelimits to count together.
        Basically to share the same ratelimit across one or more RPCs.
    :param keys: Client/s identifier. Like `group` but is used to group calls based on gRPCs message and context.
        For example by clients user-agent, ip and other requests information.
    """

    def decorator(fn):
        @wraps(fn)
        def _wrapped(self, request, context):
            current_calls, time_left = record_call(fn, request, context, time_period, group, keys)

            if current_calls > max_calls:
                details = (f"Reached limit of {max_calls} calls per {time_period} seconds."
                           f" Resource will be available in {time_left} seconds.")
                context.abort(grpc.StatusCode.RESOURCE_EXHAUSTED, details)

            return fn(self, request, context)

        return _wrapped

    return decorator



from abc import ABCMeta
import logging
from concurrent import futures

import grpc
from django.utils.module_loading import import_string

from django_grpc.signals.wrapper import SignalWrapper


logger = logging.getLogger(__name__)


def create_server(max_workers, port, interceptors=None):
    from django.conf import settings
    config = getattr(settings, 'GRPCSERVER', dict())
    servicers_list = config.get('servicers', [])  # callbacks to add servicers to the server
    interceptors = load_interceptors(config.get('interceptors', []))
    maximum_concurrent_rpcs = config.get('maximum_concurrent_rpcs', None)
    options = config.get('options', [])
    credentials = config.get('credentials', None)

    # create a gRPC server
    server = grpc.server(
        thread_pool=futures.ThreadPoolExecutor(max_workers=max_workers),
        interceptors=interceptors,
        maximum_concurrent_rpcs=maximum_concurrent_rpcs,
        options=options
    )
    add_servicers(server, servicers_list)

    if credentials is None:
        server.add_insecure_port('[::]:%s' % port)
    else:
        credential_data = list()
        for credential in credentials:
            # read in key and certificate
            with open(credential.get('private_key'), 'rb') as pp:
                private_key = pp.read()
            with open(credential.get('certificate_chain'), 'rb') as cp:
                certificate_chain = cp.read()

            credential_data.append((private_key, certificate_chain,))

        # create server credentials
        logger.debug("Adding server credentials...")
        server_credentials = grpc.ssl_server_credentials(credential_data)

        # add secure port with credentials
        server.add_secure_port('[::]:%s' % port, server_credentials)

    return server


def add_servicers(server, servicers_list):
    """
    Add servicers to the server
    """
    ps = SignalWrapper(server)
    if len(servicers_list) == 0:
        logger.warning("No servicers configured. Did you add GRPSERVER['servicers'] list to settings?")

    for path in servicers_list:
        logger.debug("Adding servicers from %s", path)
        callback = import_string(path)
        callback(ps)


def load_interceptors(strings) -> tuple:
    # Default interceptors
    result = []
    # User defined interceptors
    for path in strings:
        logger.debug("Initializing interceptor from %s", path)
        result.append(import_string(path)())
    return tuple(result)


def extract_handlers(server):
    for handler in server._state.generic_handlers:
        for path, it in handler._method_handlers.items():
            unary = it.unary_unary
            if unary is None:
                name = "???"
                params = "???"
                abstract = 'DOES NOT EXIST'
            else:
                code = it.unary_unary.__code__
                name = code.co_name
                params = ", ".join(code.co_varnames)
                abstract = ''
                if isinstance(it.__class__, ABCMeta):
                    abstract = 'NOT IMPLEMENTED'

            yield "{path}: {name}({params}) {abstract}".format(
                path=path,
                name=name,
                params=params,
                abstract=abstract
            )
